use crate::model::{
    auth::User,
    layouts::{Layout, LayoutPage, LayoutPrivacy},
    permissions::FinePermission,
    Error, Result,
};
use crate::{auto_method, DataManager};
use oiseau::{PostgresRow, execute, get, query_rows, params, cache::Cache};

impl DataManager {
    /// Get a [`Layout`] from an SQL row.
    pub(crate) fn get_layout_from_row(x: &PostgresRow) -> Layout {
        Layout {
            id: get!(x->0(i64)) as usize,
            created: get!(x->1(i64)) as usize,
            owner: get!(x->2(i64)) as usize,
            title: get!(x->3(String)),
            privacy: serde_json::from_str(&get!(x->4(String))).unwrap(),
            pages: serde_json::from_str(&get!(x->5(String))).unwrap(),
            replaces: serde_json::from_str(&get!(x->6(String))).unwrap(),
        }
    }

    auto_method!(get_layout_by_id(usize as i64)@get_layout_from_row -> "SELECT * FROM layouts WHERE id = $1" --name="layout" --returns=Layout --cache-key-tmpl="atto.layout:{}");

    /// Get all layouts by user.
    ///
    /// # Arguments
    /// * `id` - the ID of the user to fetch layouts for
    pub async fn get_layouts_by_user(&self, id: usize) -> Result<Vec<Layout>> {
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = query_rows!(
            &conn,
            "SELECT * FROM layouts WHERE owner = $1 ORDER BY created DESC",
            &[&(id as i64)],
            |x| { Self::get_layout_from_row(x) }
        );

        if res.is_err() {
            return Err(Error::GeneralNotFound("layout".to_string()));
        }

        Ok(res.unwrap())
    }

    /// Create a new layout in the database.
    ///
    /// # Arguments
    /// * `data` - a mock [`Layout`] object to insert
    pub async fn create_layout(&self, data: Layout) -> Result<Layout> {
        // check values
        if data.title.len() < 2 {
            return Err(Error::DataTooShort("title".to_string()));
        } else if data.title.len() > 32 {
            return Err(Error::DataTooLong("title".to_string()));
        }

        // ...
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(
            &conn,
            "INSERT INTO layouts VALUES ($1, $2, $3, $4, $5, $6, $7)",
            params![
                &(data.id as i64),
                &(data.created as i64),
                &(data.owner as i64),
                &data.title,
                &serde_json::to_string(&data.privacy).unwrap(),
                &serde_json::to_string(&data.pages).unwrap(),
                &serde_json::to_string(&data.replaces).unwrap(),
            ]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        Ok(data)
    }

    pub async fn delete_layout(&self, id: usize, user: &User) -> Result<()> {
        let layout = self.get_layout_by_id(id).await?;

        // check user permission
        if user.id != layout.owner && !user.permissions.check(FinePermission::MANAGE_STACKS) {
            return Err(Error::NotAllowed);
        }

        // ...
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(&conn, "DELETE FROM layouts WHERE id = $1", &[&(id as i64)]);

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        // ...
        self.0.1.remove(format!("atto.layout:{}", id)).await;
        Ok(())
    }

    auto_method!(update_layout_title(&str)@get_layout_by_id:MANAGE_USERS -> "UPDATE layouts SET title = $1 WHERE id = $2" --cache-key-tmpl="atto.layout:{}");
    auto_method!(update_layout_privacy(LayoutPrivacy)@get_layout_by_id:MANAGE_USERS -> "UPDATE layouts SET privacy = $1 WHERE id = $2" --serde --cache-key-tmpl="atto.layout:{}");
    auto_method!(update_layout_pages(Vec<LayoutPage>)@get_layout_by_id:MANAGE_USERS -> "UPDATE layouts SET pages = $1 WHERE id = $2" --serde --cache-key-tmpl="atto.layout:{}");
}
